#include "thread/Timer.h"
#include "Log.h"

static LinkedList<Scheduler::Timer> g_timer_queue;

namespace Scheduler
{
    void add_to_timer_queue(Timer* timer)
    {
        timer->delta_ticks = timer->total_ticks;

        for (auto* t : g_timer_queue)
        {
            if (timer->delta_ticks <= t->delta_ticks)
            {
                t->delta_ticks -= timer->delta_ticks;
                g_timer_queue.add_before(t, timer);
                return;
            }
            timer->delta_ticks -= t->delta_ticks;
        }

        g_timer_queue.append(timer);
    }

    void remove_from_timer_queue(Timer* timer)
    {
        auto maybe_next = g_timer_queue.next(timer);
        if (maybe_next.has_value())
        {
            auto next = maybe_next.value();
            next->delta_ticks += timer->delta_ticks;
        }
        g_timer_queue.remove(timer);
    }

    void tick_queue()
    {
        auto maybe_first = g_timer_queue.first();
        if (!maybe_first.has_value()) return;

        auto first = *maybe_first;
        first->delta_ticks--;

        LinkedList<Scheduler::Timer> timers_to_be_restarted;

        g_timer_queue.delayed_for_each([&](Timer* t) {
            if (t->delta_ticks == 0)
            {
                g_timer_queue.remove(t);
                t->thread->send_signal(t->signo);
                if (t->restart) timers_to_be_restarted.append(t);
                return true;
            }

            return false;
        });

        timers_to_be_restarted.consume([](Timer* t) { add_to_timer_queue(t); });
    }

    u64 Timer::ticks_left()
    {
        u64 total = 0;
        if (delta_ticks == 0) return 0;
        for (auto* t : g_timer_queue)
        {
            total += t->delta_ticks;
            if (t == this) break;
        }
        return total;
    }
}
